@@ -24,21 +24,25 @@ import yadtshell.metalogic
 import yadtshell.helper
 import yadtshell.update
 import yadtshell.uri
-from yadtshell.actionmanager import ActionManager
+from yadtshell.actionmanager import ActionManager  # NOQA
 
 import yadtshell.twisted
-import yadtshell.defer
+import yadtshell.defer  # NOQA
 
-from yadtshell.status import status
-from yadtshell.info import info
-from yadtshell.dump import dump
+import yadtshell.status as _status  # NOQA
+import yadtshell.info as _info  # NOQA
+
+from yadtshell.status import status  # NOQA
+from yadtshell.info import info  # NOQA
+from yadtshell.dump import dump  # NOQA
 
 
-VERSION = '1.5.2'
+VERSION = '1.7.1'
 
 
 def decode1364(encoded_string):
     return encoded_string.decode('rot13').decode('base64')
 
+
 def initialize_logging():
-    pass
\ No newline at end of file
+    pass
