@@ -16,7 +16,9 @@
 #   You should have received a copy of the GNU General Public License
 #   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
+
 class ActionException(Exception):
+
     def __init__(self, message, exitcode, rootcause=None):
         self.message = message
         self.exitcode = exitcode
@@ -24,7 +26,9 @@ class ActionException(Exception):
         self.args = [message]
         self.orig_protocol = None
 
+
 class TargetState(object):
+
     def __init__(self, uri, attr, target_value):
         self.uri = uri
         self.attr = attr
@@ -40,10 +44,13 @@ class TargetState(object):
         indent = ' ' * depth * 4
         return indent + prefix + '%(attr)s of %(uri)s is "%(target_value)s"\n' % vars(self)
 
+
 class State(object):
     PENDING, RUNNING, FINISHED = ['PENDING', 'RUNNING', 'FINISHED']
 
+
 class Action(object):
+
     def __init__(self, cmd, uri, attr=None, target_value=None, preconditions=None, args=None, kwargs=None):
         self.cmd = cmd
         self.uri = uri
@@ -56,10 +63,10 @@ class Action(object):
         self.executed = False
         self.name = '%s %s' % (cmd, uri)
         self.state = State.PENDING
-        self.args = args
-        self.kwargs = kwargs
+        self.args = args if args else []
+        self.kwargs = kwargs if kwargs else {}
         self.rank = self.uri
-        
+
     def are_all_preconditions_met(self, components):
         for precondition in self.preconditions:
             if not precondition.is_reached(components):
@@ -69,18 +76,22 @@ class Action(object):
     def __str__(self):
         return self.dump(include_preconditions=False)
 
-    def dump(self, depth=0, include_preconditions=True):
+    def dump(self, depth=0, include_preconditions=True, include_target_value=True):
         indent = ' ' * depth * 4
         text = indent
         text += '%(cmd)s the %(uri)s' % vars(self)
-        if self.attr and self.target_value:
-            text += ', set %(attr)s to "%(target_value)s"' % vars(self)
-        text += '\n'
+        if include_target_value:
+            if self.attr and self.target_value:
+                text += ', set %(attr)s to "%(target_value)s"' % vars(self)
+        aux_text = [key for key, value in self.kwargs.iteritems() if value]
+        if aux_text:
+            text += " (%s)" % " ".join(aux_text)
         if include_preconditions:
-            if self.args:
-                text += indent + '    args: ' + ', '.join(self.args) + '\n'
-            if self.kwargs:
-                text += indent + '    kwargs: ' + ', '.join(['%s: %s' % (key, value) for key, value in self.kwargs.iteritems()]) + '\n'
+            text += '\n'
+#            if self.args:
+#                text += indent + '    args: ' + ', '.join(self.args) + '\n'
+#            if self.kwargs:
+#                text += indent + '    kwargs: ' + ', '.join(['%s: %s' % (key, value) for key, value in self.kwargs.iteritems()]) + '\n'
             for precondition in self.preconditions:
                 text += precondition.dump(depth + 1, 'when ')
             #text += '\n'
@@ -95,7 +106,9 @@ class Action(object):
     def __lt__(self, other):
         return self.rank < other.rank
 
+
 class ActionPlan(object):
+
     def __init__(self, name, actions, nr_workers=None, nr_errors_tolerated=0):
         self.name = name
         if isinstance(actions, list):
@@ -103,13 +116,13 @@ class ActionPlan(object):
         else:
             self.actions = tuple(sorted(actions))
         if self.actions:
-            self.rank = self.actions[0].rank  
+            self.rank = self.actions[0].rank
         else:
             # TODO: yes, plans without actions feel and fail miserably
             self.rank = -1
         self.nr_workers = nr_workers
         self.nr_errors_tolerated = nr_errors_tolerated
-        
+
     def __str__(self):
         return self.dump(include_preconditions=False)
 
@@ -122,14 +135,14 @@ class ActionPlan(object):
         else:
             workers_str = ', workers *undefined*'
         return '%i items%s, %s errors tolerated' % (len(self.actions), workers_str, str(self.nr_errors_tolerated))
-        
+
     def dump(self, depth=0, include_preconditions=True):
         indent = ' ' * depth * 4
         text = '%s%s [%s]:\n' % (indent, self.name, self.meta_info())
         for action in self.actions:
             text += action.dump(depth + 1, include_preconditions)
         return text
-    
+
     def __lt__(self, other):
         return self.rank < other.rank
 
@@ -143,10 +156,31 @@ class ActionPlan(object):
                 if action.name == name:
                     return action.search(rest)
         return None
-    
+
     def list_subplans(self):
         yield (self.name, self)
         for plan in [p for p in self.actions if isinstance(p, ActionPlan)]:
             for sp in plan.list_subplans():
                 yield ('%s/%s' % (self.name, sp[0]), sp[1])
-                
+
+    @property
+    def list_actions(self):
+        for plan_or_action in self.actions:
+            if isinstance(plan_or_action, ActionPlan):
+                for a in plan_or_action.list_actions:
+                    yield a
+            else:
+                yield plan_or_action
+
+    @property
+    def is_empty(self):
+        return len(self.actions) == 0
+
+    @property
+    def is_not_empty(self):
+        return not self.is_empty
+
+    def remove_actions_on_unhandled_hosts(self, handled_hosts, components):
+        actions_on_handled_hosts = [
+            action for action in self.actions if components[action.uri].host_uri in handled_hosts]
+        self.actions = tuple(sorted(actions_on_handled_hosts))
