@@ -23,7 +23,11 @@ import os
 import subprocess
 import sys
 import yaml
+import shlex
 
+from twisted.internet import defer, reactor, task
+
+from yadtshell.util import calculate_max_tries_for_interval_and_delay
 import yadtshell
 
 
@@ -31,21 +35,24 @@ logger = logging.getLogger('components')
 
 
 class Component(object):
+
     def __init__(self, t, host, name=None, version=None):
         self.type = t
-        if isinstance(host, str):
+        if type(host) in [str, unicode]:
             self.host = host
         else:
             self.host = host.host
             self.fqdn = host.fqdn
-        self.host_uri = yadtshell.uri.create(yadtshell.settings.HOST, self.host)
+        self.host_uri = yadtshell.uri.create(
+            yadtshell.settings.HOST, self.host)
         self.name = getattr(self, 'name', name)
         if self.name is not None:
             self.name = self.name.rstrip('/').split('/', 1)[0]
         self.version = version
         if self.version is not None:
             self.version = str(self.version)
-        self.uri = yadtshell.uri.create(self.type, self.host, self.name, self.version)
+        self.uri = yadtshell.uri.create(
+            self.type, self.host, self.name, self.version)
         self.version = yadtshell.uri.parse(self.uri)['version']
         self.name_and_version = yadtshell.uri.as_source_file(self.uri)
         self.state = yadtshell.settings.UNKNOWN
@@ -84,7 +91,7 @@ class Component(object):
     def remote_call(self, cmd, tag=None, force=False):
         if not cmd:
             return
-        if type(cmd) is not str:
+        if type(cmd) not in [str, unicode]:
             cmd = '\n'.join(cmd)
 
         ssh_cmd = yadtshell.settings.SSH
@@ -98,8 +105,8 @@ class Component(object):
         remotecall_script = '/usr/bin/yadt-remotecall'
         log_file = self.create_remote_log_filename(tag=tag)
         owner = yadtshell.util.get_locking_user_info()['owner']
-        is_force = {False: '', True: '-f'}[force]
-        complete_cmd = '''%(ssh_cmd)s %(host)s %(remotecall_script)s -l %(log_file)s -m %(owner)s -s %(service)s %(is_force)s \"%(cmd)s\" ''' % locals()
+        is_force = {False: '', True: ' --force'}[force]
+        complete_cmd = '%(ssh_cmd)s %(host)s WHO="%(owner)s" YADT_LOG_FILE="%(log_file)s" "yadt-command %(cmd)s%(is_force)s" ' % locals()
         return complete_cmd
 
     def local_call(self, cmd, tag=None, guard=True, force=False, no_subprocess=True):
@@ -113,7 +120,9 @@ class Component(object):
         if no_subprocess:
             return cmds
         if guard:
-            sp = self.remote_call(": #check service callable", tag='check_service_callable', force=force)
+            sp = self.remote_call(
+                ": #check service callable", tag='check_service_callable',
+                force=force)
             returncode = yadtshell.util.log_subprocess(sp)
             if returncode != 0:
                 return returncode
@@ -131,7 +140,8 @@ class Component(object):
     def _create_owner_file(self, lockinfo, filename, force=False, tag=None):
         """@return: integer The error code of the remote call"""
         dirname = os.path.dirname(filename)
-        cmd = '''umask 0002 && mkdir -pv %s && echo -e '%s' > %s''' % (dirname, yadtshell.util.get_yaml(lockinfo), filename)
+        cmd = '''umask 0002 && mkdir -pv %s && echo -e '%s' > %s''' % (
+            dirname, yadtshell.util.get_yaml(lockinfo), filename)
         return self.remote_call(cmd, tag, force=force)
 
     def _remove_owner_file(self, lockinfo, filename, force=False, tag=None):
@@ -139,13 +149,18 @@ class Component(object):
         cmd = "rm -fv %(filename)s" % locals()
         return self.remote_call(cmd, tag, force=force)
 
+
 class MissingComponent(Component):
+
     def __init__(self, s):
         parts = yadtshell.uri.parse(s)
-        Component.__init__(self, parts['type'], parts['host'], parts['name'], parts['version'])
+        Component.__init__(self, parts['type'], parts[
+                           'host'], parts['name'], parts['version'])
         self.state = yadtshell.settings.MISSING
 
+
 class ComponentDict(dict):
+
     def __init__(self):
         dict.__init__(self)
         self._add_when_missing_ = False
@@ -161,16 +176,21 @@ class ComponentDict(dict):
             logger.debug('missing ' + key)
             self[self._key_(key)] = MissingComponent(self._key_(key))
         return dict.__getitem__(self, self._key_(key))
+
     def get(self, key, default=None):
         if self._key_(key) not in self and self._add_when_missing_:
             logger.debug('missing' + key)
             self[self._key_(key)] = MissingComponent(key)
         return dict.get(self, self._key_(key), default)
+
     def __setitem__(self, key, value):
-        #logger.debug('adding ' + self._key_(key) + ' ' + getattr(value, 'state', ''))
+        # logger.debug('adding ' + self._key_(key) + ' ' + getattr(value,
+        # 'state', ''))
         return dict.__setitem__(self, self._key_(key), value)
 
+
 class ComponentSet(set):
+
     def __init__(self, components=None):
         self.components = components
         self._set = set([])
@@ -186,7 +206,8 @@ class ComponentSet(set):
         logger.debug('adding ' + key)
         if key not in self.components and check:
             logger.warning('key %(key)s not found, ignoring' % locals())
-            #logger.warning('known keys: ' + ', '.join(self.components.keys()))
+            # logger.warning('known keys: ' + ',
+            # '.join(self.components.keys()))
             return None
         return self._set.add(key)
 
@@ -206,18 +227,91 @@ class ComponentSet(set):
     def __contains__(self, item):
         return self._key_(item) in self._set
 
+
 class Host(Component):
+
     def __init__(self, name):
         self.lockstate = None
         self.is_locked = None
         self.is_locked_by_other = None
         self.is_locked_by_me = None
+        self.ssh_poll_max_seconds = yadtshell.constants.SSH_POLL_MAX_SECONDS_DEFAULT
+        self.reboot_required_to_activate_latest_kernel = False
+        self.reboot_required_after_next_update = False
         Component.__init__(self, yadtshell.settings.HOST, name)
 
-    def update(self):
-        return self.remote_call('sudo /usr/bin/yadt-yum upgrade', '%s_%s' %
+    @property
+    def reboot_required(self):
+        return self.reboot_required_after_next_update or self.reboot_required_to_activate_latest_kernel
+
+    def is_reachable(self):
+        return True
+
+    def update(self, reboot_required=False):
+        next_artefacts = [uri.replace('/', '-', 1)
+                          for uri in self.next_artefacts]
+        if not reboot_required:
+            return self.remote_call(
+                'yadt-host-update %s' % ' '.join(next_artefacts), '%s_%s' %
                 (self.hostname, yadtshell.settings.UPDATE))
 
+        update_and_reboot_command = self.remote_call(
+            'yadt-host-update -r %s' % ' '.join(next_artefacts), '%s_%s' %
+            (self.hostname, yadtshell.settings.UPDATE))
+        p = yadtshell.twisted.YadtProcessProtocol(self,
+                                                  update_and_reboot_command,
+                                                  out_log_level=logging.INFO)
+        p.target_state = yadtshell.settings.UPTODATE
+        p.state = yadtshell.settings.UNKNOWN
+
+        p.deferred = defer.Deferred()
+
+        def handle_rebooting_machine(failure, ssh_poll_max_seconds):
+            if failure.value.exitCode == 152:
+                raise yadtshell.actions.ActionException(
+                    'Timed out while waiting for %s to reboot' % self.uri)
+            elif failure.value.exitCode == 255:
+                logger.info("%s: rebooting now" % self.uri)
+                return poll_rebooting_machine(ssh_poll_max_seconds=ssh_poll_max_seconds)
+            return failure
+
+        def poll_rebooting_machine(count=1, ssh_poll_max_seconds=yadtshell.constants.SSH_POLL_MAX_SECONDS_DEFAULT):
+            max_tries = calculate_max_tries_for_interval_and_delay(interval=ssh_poll_max_seconds,
+                                                                   delay=yadtshell.constants.SSH_POLL_DELAY)
+            logger.info("%s: polling for ssh connect, try %i of %i" %
+                        (self.uri, count, max_tries))
+            poll_command = self.remote_call(
+                'uptime', '%s_poll' % self.hostname)
+            poll_protocol = yadtshell.twisted.YadtProcessProtocol(
+                self, poll_command, out_log_level=logging.INFO)
+            poll_protocol.ssh_poll_count = count
+            poll_protocol.deferred = defer.Deferred()
+            if (count * yadtshell.constants.SSH_POLL_DELAY) < ssh_poll_max_seconds:
+                poll_protocol.deferred.addErrback(lambda x:
+                                                  task.deferLater(reactor,
+                                                                  yadtshell.constants.SSH_POLL_DELAY,
+                                                                  poll_rebooting_machine,
+                                                                  count + 1,
+                                                                  ssh_poll_max_seconds=ssh_poll_max_seconds))
+            cmdline = shlex.split(poll_protocol.cmd)
+            reactor.spawnProcess(poll_protocol, cmdline[0], cmdline, None)
+
+            return poll_protocol.deferred
+
+        p.deferred.addErrback(
+            handle_rebooting_machine, self.ssh_poll_max_seconds)
+
+        def display_reboot_info(protocol):
+            if hasattr(protocol, 'ssh_poll_count'):
+                logger.info('%s: reboot took %d seconds' %
+                            (self.uri, protocol.ssh_poll_count * yadtshell.constants.SSH_POLL_DELAY))
+            return protocol
+        p.deferred.addCallback(display_reboot_info)
+
+        cmdline = shlex.split(p.cmd.encode('ascii'))
+        reactor.spawnProcess(p, cmdline[0], cmdline, None)
+        return p.deferred
+
     def bootstrap(self):
         pass    # TODO to be implemented
 
@@ -228,10 +322,10 @@ class Host(Component):
         return self.state == yadtshell.settings.UPDATE_NEEDED
 
     def probe(self):
-        return self.remote_call('/usr/bin/yadt-status-host')
+        return self.remote_call('yadt-host-status')
 
     def probe_uptodate(self):
-        return yadtshell.util.log_subprocess(self.remote_call('/usr/bin/yadt-status-host', '%s_probe' % self.hostname))
+        return yadtshell.util.log_subprocess(self.remote_call('yadt-host-status', '%s_probe' % self.hostname))
 
     def get_lock_dir(self):
         return self.defaults['YADT_LOCK_DIR']
@@ -246,23 +340,20 @@ class Host(Component):
             return None
 
     def lock(self, message=None, force=False, **kwargs):
+        def strip_quotes_from_message(message):
+            return message.replace("'", "").replace('"', '')
         if not message:
             raise ValueError('the "message" parameter is mandatory')
         lockinfo = yadtshell.util.get_locking_user_info()
         lockinfo["message"] = message
         lockinfo["force"] = force
-        return self._create_owner_file(
-                lockinfo,
-                os.path.join(self.get_lock_dir(), 'host.lock'),  # TODO extract method for filename
-                force=force, tag="lock_host")
+        return self.remote_call(
+            "yadt-host-lock '%s'" % strip_quotes_from_message(message),
+            'lock_host',
+            force)
 
     def unlock(self, force=False, **kwargs):
-        lockinfo = yadtshell.util.get_locking_user_info()
-        return self._remove_owner_file(
-                lockinfo,
-                os.path.join(self.get_lock_dir(), 'host.lock'),  # TODO extract method for filename
-                force=force,
-                tag="unlock_host")
+        return self.remote_call('yadt-host-unlock', "unlock_host", force)
 
     def update_attributes_after_status(self):
         self.is_locked = not self.lockstate is None
@@ -271,26 +362,56 @@ class Host(Component):
         lock_owner = None
         if self.lockstate:
             lock_owner = self.lockstate.get("owner")
-        self.is_locked_by_me = self.is_locked and lock_owner and lock_owner == lockinfo["owner"]
+        self.is_locked_by_me = self.is_locked and lock_owner and lock_owner == lockinfo[
+            "owner"]
 
         self.is_locked_by_other = self.is_locked and not self.is_locked_by_me
 
-        logger.debug("is_locked=" + repr(self.is_locked) + ", is_locked_by_me=" + repr(self.is_locked_by_me) + ", is_locked_by_other=" + repr(self.is_locked_by_other))
+        logger.debug("is_locked=" + repr(self.is_locked) + ", is_locked_by_me=" + repr(
+            self.is_locked_by_me) + ", is_locked_by_other=" + repr(self.is_locked_by_other))
+
+
+class UnreachableHost(Component):
+
+    def __init__(self, name):
+        self.fqdn = name
+        self.hostname = self.fqdn.split('.')[0]
+        Component.__init__(self, yadtshell.settings.HOST, name)
+
+    def is_reachable(self):
+        return False
+
+    def is_unknown(self):
+        return True
+
+    @property
+    def next_artefacts(self):
+        return {}
+
+    @property
+    def is_locked_by_other(self):
+        return False
+
+    @property
+    def is_locked_by_me(self):
+        return False
 
 
 class Artefact(Component):
+
     def __init__(self, host, name, version=None):
-        Component.__init__(self, yadtshell.settings.ARTEFACT, host, name, version)
-        #self.needs.add(uri.create(settings.HOST, host.host))
+        Component.__init__(
+            self, yadtshell.settings.ARTEFACT, host, name, version)
+        # self.needs.add(uri.create(settings.HOST, host.host))
 
     def updateartefact(self):
-        return self.remote_call('sudo /usr/bin/yadt-yum upgrade -y %s' % self.name,
+        return self.remote_call('yadt-artefact-update %s' % self.name,
                                 'artefact_%s_%s_%s' % (self.host, self.name, yadtshell.constants.UPDATEARTEFACT))
 
 
-
 class Service(Component):
-    def __init__(self, host, name, settings = None):
+
+    def __init__(self, host, name, settings=None):
         Component.__init__(self, yadtshell.settings.SERVICE, host, name)
 
         self.needs_services = []
@@ -310,63 +431,52 @@ class Service(Component):
             if n.startswith(yadtshell.settings.SERVICE):
                 self.needs.add(n % locals())
             else:
-                self.needs.add(yadtshell.uri.create(yadtshell.settings.SERVICE, host.host, n % locals()))
+                self.needs.add(yadtshell.uri.create(
+                    yadtshell.settings.SERVICE, host.host, n % locals()))
         for n in self.needs_artefacts:
             self.needs.add(yadtshell.uri.create(yadtshell.settings.ARTEFACT, host.host, n % locals() +
-                "/" + yadtshell.settings.CURRENT))
-        #self.needs.add(uri.create(yadtshell.settings.HOST, host.host))
+                                                "/" + yadtshell.settings.CURRENT))
+        # self.needs.add(uri.create(yadtshell.settings.HOST, host.host))
 
-        self.state = yadtshell.settings.STATE_DESCRIPTIONS.get(settings.get('state'),
-                yadtshell.settings.UNKNOWN)
+        self.state = yadtshell.settings.STATE_DESCRIPTIONS.get(
+            settings.get('state'),
+            yadtshell.settings.UNKNOWN)
         self.script = None
 
-
     def stop(self, force=False, **kwargs):
-        return self.remote_call(self._retrieve_service_call(yadtshell.settings.STOP),
-                '%s_%s' % (self.name, yadtshell.settings.STOP), force)
+        return self.remote_call(
+            self._retrieve_service_call(yadtshell.settings.STOP),
+            '%s_%s' % (self.name, yadtshell.settings.STOP), force)
 
     def start(self, force=False, **kwargs):
-        return self.remote_call(self._retrieve_service_call(yadtshell.settings.START),
-                '%s_%s' % (self.name, yadtshell.settings.START), force)
+        return self.remote_call(
+            self._retrieve_service_call(yadtshell.settings.START),
+            '%s_%s' % (self.name, yadtshell.settings.START), force)
 
     def status(self):
-        return self.remote_call(self._retrieve_service_call(yadtshell.settings.STATUS),
-                tag='%s_%s' % (self.name, yadtshell.settings.STATUS), force=True)
+        return self.remote_call(
+            self._retrieve_service_call(yadtshell.settings.STATUS),
+            tag='%s_%s' % (self.name, yadtshell.settings.STATUS))
 
     def _retrieve_service_call(self, action):
-        name = self.name
-        return 'sudo /sbin/service %(name)s %(action)s' % locals()
+        return 'yadt-service-%s %s' % (action, self.name)
 
     def ignore(self, message=None, **kwargs):
         if not message:
             raise ValueError('the "message" parameter is mandatory')
-        components = yadtshell.util.restore_current_state()
-        host = components[self.host_uri]
-        result = self._create_owner_file(
-            yadtshell.util.get_locking_user_info(),
-            os.path.join(host.get_ignored_dir(), 'ignore.%s' % self.name),  # TODO extract method for filename
-            force=kwargs.get('force', False),
-            tag="ignore_%s" % self.name)
-        if not result:
-            logger.warn("Could not ignore %s. Try --force" % self.uri)
-        return result
+        tag = "ignore_%s" % self.name
+        force = kwargs.get('force', False)
+        return self.remote_call('yadt-service-ignore %s "%s"' % (self.name, message), tag, force)
 
     def unignore(self, **kwargs):
-        components = yadtshell.util.restore_current_state()
-        host = components[self.host_uri]
-        result = self._remove_owner_file(
-            yadtshell.util.get_locking_user_info(),
-            os.path.join(host.get_ignored_dir(), 'ignore.%s' % self.name),  # TODO extract method for filename
-            force=True, #kwargs.get('force'),
-            tag="unignore_%s" % self.name)
-        if not result:
-            logger.warn("Could not unignore %s. Try --force" % self.uri)
-        return result
+        tag = "unignore_%s" % self.name
+        return self.remote_call('yadt-service-unignore %s' % self.name, tag)
 
 
 def do_cb(protocol, args, opts):
     return do(args, opts)
 
+
 def do(args, opts):
     cmd = args[0]
     component_names = args[1:]
@@ -381,7 +491,8 @@ def do(args, opts):
     for component_name in component_names:
         component = components.get(component_name, None)
         if not component:
-            component = components[yadtshell.uri.change_version(component_name, 'current')]
+            component = components[
+                yadtshell.uri.change_version(component_name, 'current')]
         fun = getattr(component, cmd, None)
         import inspect
         if inspect.ismethod(fun):
@@ -390,16 +501,19 @@ def do(args, opts):
             except TypeError:
                 sp = fun()
         else:
-            logger.error('"%(cmd)s" is not defined for %(component_name)s, aborting' % locals())
+            logger.error(
+                '"%(cmd)s" is not defined for %(component_name)s, aborting' % locals())
             sys.exit(2)
         logger.debug('%(cmd)sing %(component_name)s' % locals())
         try:
             logger.debug('executing fun ' + str(fun))
             if isinstance(sp, subprocess.Popen):
-                exit_code = yadtshell.util.log_subprocess(sp, stdout_level=logging.INFO)
+                exit_code = yadtshell.util.log_subprocess(
+                    sp, stdout_level=logging.INFO)
             else:
                 exit_code = sp
             logger.debug('exit code %(exit_code)s' % locals())
         except AttributeError, ae:
-            logger.warning('problem while executing %(cmd)s on %(component_name)s' % locals())
+            logger.warning(
+                'problem while executing %(cmd)s on %(component_name)s' % locals())
             logger.exception(ae)
