@@ -22,11 +22,17 @@ PROBED = 'probed'
 UPDATEARTEFACT = 'updateartefact'
 
 UPDATE_NEEDED = 'update_needed'
+REBOOT_REQUIRED = 'reboot_required'
 
 HOST_STATE_DESCRIPTIONS = {
     0: 'uptodate',
     1: UPDATE_NEEDED
-} 
+}
 
 IGNORE = 'ignore'
-UNIGNORE = 'unignore'
\ No newline at end of file
+UNIGNORE = 'unignore'
+
+STANDALONE_SERVICE_RANK = 49152
+
+SSH_POLL_DELAY = 5
+SSH_POLL_MAX_SECONDS_DEFAULT = 120
