@@ -30,9 +30,11 @@ import yadtshell.TerminalController
 from yadtshell.helper import condense_hosts, condense_hosts2, get_user_info
 from yadtshell.loggingtools import create_next_log_file_name_with_command_arguments_as_tag
 import yadtshell.helper
+from yadtshell.loggingtools import configure_logger_output_stream_by_level
 
 sys.path.append('/etc/yadtshell')
 
+DEFAULT_LOG_DIR = '/tmp/logs/yadtshell'
 USER_INFO = get_user_info()
 OUTPUT_DIR = os.path.expanduser('~%s/.yadtshell' % USER_INFO['user'])
 
@@ -42,32 +44,44 @@ root_logger = logging.getLogger()
 root_logger.setLevel(logging.DEBUG)
 
 message_formatter = logging.Formatter('%(levelname)-8s %(message)s')
-console_handler = logging.StreamHandler()
-console_handler.setLevel(logging.INFO)
-console_handler.setFormatter(message_formatter)
-root_logger.addHandler(console_handler)
+# DO NOT USE A KWARG : it's 'strm' in python<2.6 and has
+# been renamed to 'stream' in 2.7 with NO DOCUMENTATION...
+console_stdout_handler = logging.StreamHandler(sys.stdout)
+console_stderr_handler = logging.StreamHandler(sys.stderr)
+configure_logger_output_stream_by_level(
+    console_stderr_handler, console_stdout_handler)
+console_stdout_handler.setFormatter(message_formatter)
+console_stderr_handler.setFormatter(message_formatter)
+root_logger.addHandler(console_stdout_handler)
+root_logger.addHandler(console_stderr_handler)
+
+
+class SettingsError(BaseException):
+    pass
 
 logger = logging.getLogger('settings')
 
-
 try:
-    from loggingconf import *
+    from loggingconf import *  # NOQA
 except Exception, e:
     root_logger.debug(e)
-    LOG_DIR_PREFIX = '/var/log/yadtshell'
+    if not os.path.isdir(DEFAULT_LOG_DIR):
+        os.makedirs(DEFAULT_LOG_DIR)  # pragma: no cover
+    LOG_DIR_PREFIX = DEFAULT_LOG_DIR
 
 
-def initialize_broadcast_client():
+def initialize_broadcast_client():  # pragma: no cover
     global DummyBroadcaster, broadcasterconf_imported, broadcasterconf, e
 
     class DummyBroadcaster(object):
+
         def addOnSessionOpenHandler(self, *args, **kwargs):
             pass
 
-        def sendServiceChange(self, data):
+        def sendServiceChange(self, data, **kwargs):
             pass
 
-        def sendFullUpdate(self, data):
+        def sendFullUpdate(self, data, **kwargs):
             pass
 
         def connect(self):
@@ -79,7 +93,7 @@ def initialize_broadcast_client():
     broadcasterconf_imported = False
     try:
         sys.path.append("/etc/yadtbroadcast-client/")
-        import broadcasterconf
+        import broadcasterconf  # NOQA
         sys.path.pop()
         broadcasterconf_imported = True
     except Exception, e:
@@ -87,7 +101,59 @@ def initialize_broadcast_client():
         logger.warn(e)
 
 
-def load_settings():
+def _load_target_file(target_filename, targets_dir, target_basename, visited=None):
+    if not visited:
+        visited = []
+
+    if target_filename in visited:
+        return {}
+
+    visited.append(target_filename)
+
+    try:
+        settings_file = open(target_filename)
+    except EnvironmentError:
+        raise SettingsError(
+            'cannot find target definition file %s, aborting' % target_filename)
+    target_settings = yaml.load(settings_file)
+    settings_file.close()
+
+    if target_settings.get('hosts') is None:  # hosts:\n means hosts==None in YAML
+        target_settings['hosts'] = []         # so either the key is missing OR hosts is a null-value
+    for include in target_settings.get('includes', []):
+        subtarget_filename = os.path.join(
+            targets_dir, include, target_basename)
+        subtarget_settings = _load_target_file(
+            subtarget_filename, targets_dir, target_basename, visited)
+        for host in subtarget_settings.get('hosts', []):
+            target_settings['hosts'].append(host)
+    return target_settings
+
+
+def load_target_file(target_basename):
+    targets_dir = os.getenv("TARGETS_DIR", os.path.dirname(os.getcwd()))
+    target_name = os.path.basename(os.getcwd())
+    target_filename = os.path.join(targets_dir, target_name, target_basename)
+    target_settings = _load_target_file(
+        target_filename, targets_dir, target_basename)
+    target_settings.setdefault('name', target_name)
+    target_settings = expand_hosts(target_settings)
+    return target_settings
+
+
+def expand_hosts(target_settings):
+    he = hostexpand.HostExpander.HostExpander(
+        outputformat=hostexpand.HostExpander.HostExpander.FQDN)
+    target_settings['original_hosts'] = target_settings['hosts']
+    target_settings['hosts'] = he.expand(target_settings['hosts'])
+    return target_settings
+
+
+def load_settings(log_to_file=True):
+    TARGET_BASENAME = 'target'
+
+    global TARGET_SETTINGS
+    TARGET_SETTINGS = load_target_file(TARGET_BASENAME)
 
     initialize_broadcast_client()
 
@@ -112,18 +178,6 @@ def load_settings():
     global term
     term = yadtshell.TerminalController.TerminalController()
 
-    TARGET_SETTINGS_FILE = 'target'
-    try:
-        settings_file = open(TARGET_SETTINGS_FILE)
-    except IOError:
-        root_logger.critical('cannot find target definition file, aborting')
-        sys.exit(1)
-    global TARGET_SETTINGS
-    TARGET_SETTINGS = yaml.load(settings_file)
-    settings_file.close()
-
-    TARGET_SETTINGS.setdefault('name', os.path.basename(os.getcwd()))
-
     global ybc
     if broadcasterconf_imported:
         ybc = broadcasterconf.create(TARGET_SETTINGS['name'])
@@ -141,38 +195,38 @@ def load_settings():
 
     global log_file
     log_file = create_next_log_file_name_with_command_arguments_as_tag(
-                    log_dir=LOG_DIR,
-                    target_name=TARGET_SETTINGS['name'],
-                    command_start_timestamp=STARTED_ON,
-                    user_name=USER_INFO['user'],
-                    source_host=USER_INFO['yadt_host'].split('.')[0],
-                    command_arguments=sys.argv
+        log_dir=LOG_DIR,
+        target_name=TARGET_SETTINGS['name'],
+        command_start_timestamp=STARTED_ON,
+        user_name=USER_INFO['user'],
+        source_host=USER_INFO['yadt_host'].split('.')[0],
+        command_arguments=sys.argv
     )
 
-    formatter = logging.Formatter('%(asctime)s %(levelname)s %(filename)s:%(lineno)d %(message)s', '%Y%m%d-%H%M%S')
-    file_handler = logging.FileHandler(log_file)
-    file_handler.setLevel(logging.DEBUG)
-    file_handler.setFormatter(formatter)
-    root_logger.addHandler(file_handler)
+    formatter = logging.Formatter(
+        '%(asctime)s %(levelname)s %(filename)s:%(lineno)d %(message)s', '%Y%m%d-%H%M%S')
+
+    if log_to_file:
+        file_handler = logging.FileHandler(log_file)
+        file_handler.setLevel(logging.DEBUG)
+        file_handler.setFormatter(formatter)
+        root_logger.addHandler(file_handler)
 
     logger.debug(yaml.dump(USER_INFO, default_flow_style=False))
     logger.debug('Called "{0}"'.format(' '.join(sys.argv)))
     logger.debug('output dir is %s' % OUTPUT_DIR)
 
-    he = hostexpand.HostExpander.HostExpander(outputformat=hostexpand.HostExpander.HostExpander.FQDN)
-    TARGET_SETTINGS['original_hosts'] = TARGET_SETTINGS['hosts']
-    TARGET_SETTINGS['hosts'] = he.expand(TARGET_SETTINGS['hosts'])
-
-    OUT_TARGET_FILE = os.path.join(OUT_DIR, TARGET_SETTINGS_FILE)
+    OUT_TARGET_FILE = os.path.join(OUT_DIR, TARGET_BASENAME)
     try:
-        changed = not filecmp.cmp(TARGET_SETTINGS_FILE, OUT_TARGET_FILE)
+        changed = not filecmp.cmp(TARGET_BASENAME, OUT_TARGET_FILE)
     except OSError:
         changed = True
     if changed:
-        logger.info('target settings have changed since last call, thus cleaning cached data')
+        logger.info(
+            'target settings have changed since last call, thus cleaning cached data')
         shutil.rmtree(OUT_DIR)
         os.makedirs(OUT_DIR)
-        shutil.copy2(TARGET_SETTINGS_FILE, OUT_TARGET_FILE)
+        shutil.copy2(TARGET_BASENAME, OUT_TARGET_FILE)
 
     global VIEW_SETTINGS
     VIEW_SETTINGS = {'info-view': ['matrix', 'color', 'maxcols']}
@@ -183,16 +237,21 @@ def load_settings():
         view_file.close()
 
     except:
-        logger.debug('"view" file not found, falling back to default values: %s' %
-                VIEW_SETTINGS)
+        logger.debug(
+            '"view" file not found, falling back to default values: %s' %
+            VIEW_SETTINGS)
 
     hosts_condensed_file = open(os.path.join(OUT_DIR, 'hosts_condensed'), 'w')
-    print >> hosts_condensed_file, ', '.join(condense_hosts2(condense_hosts(TARGET_SETTINGS['hosts'])))
+    print >> hosts_condensed_file, ' '.join(
+        condense_hosts2(condense_hosts(TARGET_SETTINGS['hosts'])))
     hosts_condensed_file.close()
 
     def list_selected_hosts():
         return 'You are working now on %s\n\nas full list: %s\n' % (
-            term.render('${BOLD}') + ', '.join(condense_hosts2(condense_hosts(TARGET_SETTINGS['hosts']))) + term.render('${NORMAL}'),
+            term.render('${BOLD}') + ', '.join(
+                condense_hosts2(
+                    condense_hosts(
+                        TARGET_SETTINGS['hosts']))) + term.render('${NORMAL}'),
             ', '.join(TARGET_SETTINGS['hosts']),
         )
 
@@ -212,7 +271,11 @@ def load_settings():
     except OSError:
         pass
     global SSH
-    SSH = 'ssh -o ControlPath=%s -A %s -T -o ConnectTimeout=4 -o BatchMode=yes -o CheckHostIP=no -o StrictHostKeyChecking=no' % (SSH_CONTROL_PATH, credentials)
+    SSH = 'ssh -o ControlPath=%s -A %s -T -o ConnectTimeout=4 -o BatchMode=yes -o CheckHostIP=no -o StrictHostKeyChecking=no -q' % (
+        SSH_CONTROL_PATH, credentials)
+
+    global tracking_id
+    tracking_id = None
 
 
 HOST = "host"
@@ -260,4 +323,4 @@ STATE_DESCRIPTIONS = {
     UP: UP,
     DOWN: DOWN,
     UNKNOWN: UNKNOWN,
-}
\ No newline at end of file
+}
