@@ -26,6 +26,7 @@ import traceback
 import inspect
 import shlex
 import yaml
+import simplejson as json
 
 from twisted.internet import protocol
 from twisted.internet import reactor
@@ -33,6 +34,8 @@ from twisted.internet import defer
 
 from hostexpand.HostExpander import HostExpander
 import yadtshell
+from yadtshell.util import compute_dependency_scores
+
 
 logger = logging.getLogger('status')
 
@@ -43,7 +46,63 @@ def status_cb(protocol=None):
     return status()
 
 
-def status(hosts=None, include_artefacts=True, use_cache_only=False, **kwargs):
+def query_status(component, pi=None):
+    p = yadtshell.twisted.YadtProcessProtocol(
+        component, '/usr/bin/yadt-status', pi)
+    p.deferred = defer.Deferred()
+    p.deferred.name = component
+    cmd = shlex.split(yadtshell.settings.SSH) + [component]
+    reactor.spawnProcess(p, cmd[0], cmd, os.environ)
+    return p.deferred
+
+
+def create_host(protocol, components, yaml_loader):
+    if isinstance(protocol, yadtshell.components.UnreachableHost):
+        unreachable_host = protocol
+        unreachable_host_uri = yadtshell.uri.create(
+            type=yadtshell.settings.HOST, host=unreachable_host.hostname)
+        components[unreachable_host_uri] = unreachable_host
+        return unreachable_host
+
+    def convert_string_to_host(data, host=None):
+        try:
+            return json.loads(data)
+        except Exception, e:
+            logger.debug(
+                '%s: %s, falling back to yaml parser' % (host, str(e)))
+            return yaml.load(data, Loader=yaml_loader)
+
+    host = None
+    data = convert_string_to_host(protocol.data, protocol.component)
+    if data == yadtshell.settings.DOWN:
+        host = yadtshell.components.Host(protocol.component)
+        host.state = yadtshell.settings.DOWN
+    elif data == yadtshell.settings.UNKNOWN:
+        host = yadtshell.components.Host(protocol.component)
+        host.state = yadtshell.settings.UNKNOWN
+    elif data is None:
+        logging.getLogger(protocol.component).warning(
+            'no data? strange...')
+    elif "hostname" not in protocol.data:
+        logging.getLogger(protocol.component).warning(
+            'no hostname? strange...')
+    else:
+        host = yadtshell.components.Host(data['hostname'])
+        for key, value in data.iteritems():
+            setattr(host, key, value)
+        host.state = ['update_needed', 'uptodate'][not host.next_artefacts]
+    loc_type = yadtshell.util.determine_loc_type(host.hostname)
+    host.loc_type = loc_type
+    host.update_attributes_after_status()
+    host.next_artefacts = getattr(host, 'next_artefacts', [])
+    if host.next_artefacts is None:
+        host.next_artefacts = []
+    host.logger = logging.getLogger(host.uri)
+    components[host.uri] = host
+    return host
+
+
+def status(hosts=None, include_artefacts=True, **kwargs):
     try:
         from yaml import CLoader as Loader
         logger.debug("using C implementation of yaml")
@@ -61,22 +120,24 @@ def status(hosts=None, include_artefacts=True, use_cache_only=False, **kwargs):
     if type(hosts) is str:
         hosts = [hosts]
 
-    if not use_cache_only:
-        try:
-            os.remove(os.path.join(yadtshell.settings.OUT_DIR, 'current_state.components'))
-        except OSError:
-            pass
+    try:
+        os.remove(
+            os.path.join(yadtshell.settings.OUT_DIR, 'current_state.components'))
+    except OSError:
+        pass
 
-        if hosts:
-            state_files = [os.path.join(yadtshell.settings.OUT_DIR, 'current_state_%s.yaml' % h) for h in hosts]
-        else:
-            state_files = glob.glob(os.path.join(yadtshell.settings.OUT_DIR, 'current_state*'))
-        for state_file in state_files:
-            logger.debug('removing old state %(state_file)s' % locals())
-            try:
-                os.remove(state_file)
-            except OSError, e:
-                logger.warning('cannot remove %s:\n    %s' % (state_file, e))
+    if hosts:
+        state_files = [os.path.join(yadtshell.settings.OUT_DIR, 'current_state_%s.yaml' % h)
+                       for h in hosts]
+    else:
+        state_files = glob.glob(
+            os.path.join(yadtshell.settings.OUT_DIR, 'current_state*'))
+    for state_file in state_files:
+        logger.debug('removing old state %(state_file)s' % locals())
+        try:
+            os.remove(state_file)
+        except OSError, e:
+            logger.warning('cannot remove %s:\n    %s' % (state_file, e))
 
     logger.debug('starting remote queries')
 
@@ -85,65 +146,37 @@ def status(hosts=None, include_artefacts=True, use_cache_only=False, **kwargs):
 
     components = yadtshell.components.ComponentDict()
 
-    def query_status(component, pi=None):
-        p = yadtshell.twisted.YadtProcessProtocol(component, '/usr/bin/yadt-status', pi)
-        p.deferred = defer.Deferred()
-        p.deferred.name = component
-        cmd = shlex.split(yadtshell.settings.SSH) + [component]
-        reactor.spawnProcess(p, cmd[0], cmd, os.environ)
-        return p.deferred
-
-    def create_host(protocol):
-        def convert_string_to_host(data):
-            return yaml.load(data, Loader=Loader)
-
-        host = None
-        data = convert_string_to_host(protocol.data)
-        if data == yadtshell.settings.DOWN:
-            host = yadtshell.components.Host(protocol.component)
-            host.state = yadtshell.settings.DOWN
-        elif data == yadtshell.settings.UNKNOWN:
-            host = yadtshell.components.Host(protocol.component)
-            host.state = yadtshell.settings.UNKNOWN
-        elif data is None:
-            logging.getLogger(protocol.component).warning('no data? strange...')
-        elif "hostname" not in protocol.data:
-            logging.getLogger(protocol.component).warning('no hostname? strange...')
-        else:
-            host = yadtshell.components.Host(data['hostname'])
-            for key, value in data.iteritems():
-                setattr(host, key, value)
-            host.state = ['update_needed', 'uptodate'][host.next_artefacts is None]
-        loc_type = yadtshell.util.determine_loc_type(host.hostname)
-        host.loc_type = loc_type
-        host.update_attributes_after_status()
-        host.next_artefacts = getattr(host, 'next_artefacts', [])
-        if host.next_artefacts is None:
-            host.next_artefacts = []
-        host.logger = logging.getLogger(host.uri)
-        components[host.uri] = host
-        return host
-
     def store_service_up(protocol):
         protocol.component.state = yadtshell.settings.UP
         return protocol
 
     def store_service_not_up(reason):
-        reason.value.component.state = yadtshell.settings.STATE_DESCRIPTIONS.get(reason.value.exitCode, yadtshell.settings.UNKNOWN)
+        reason.value.component.state = yadtshell.settings.STATE_DESCRIPTIONS.get(
+            reason.value.exitCode, yadtshell.settings.UNKNOWN)
         return protocol
 
     def query_local_service(service):
         cmd = service.status()
         if isinstance(cmd, defer.Deferred):
-            def store_service_state(state, service):    # TODO refactor: integrate all store_service_* cbs
-                service.state = yadtshell.settings.STATE_DESCRIPTIONS.get(state, yadtshell.settings.UNKNOWN)
+            # TODO refactor: integrate all store_service_* cbs
+            def store_service_state(state, service):
+                service.state = yadtshell.settings.STATE_DESCRIPTIONS.get(
+                    state, yadtshell.settings.UNKNOWN)
             cmd.addCallback(store_service_state, service)
+
+            def handle_service_state_failure(failure, service):
+                logger.debug('Failure while determining state of {0}. Exit code was {1}.'
+                             .format(service.uri, failure.value.exitCode))
+            cmd.addErrback(handle_service_state_failure, service)
             return cmd
-        query_protocol = yadtshell.twisted.YadtProcessProtocol(service.uri, cmd)
+        query_protocol = yadtshell.twisted.YadtProcessProtocol(
+            service.uri, cmd)
         query_protocol.deferred = defer.Deferred()
-        reactor.spawnProcess(query_protocol, '/bin/sh', ['/bin/sh'], os.environ)
+        reactor.spawnProcess(
+            query_protocol, '/bin/sh', ['/bin/sh'], os.environ)
         query_protocol.component = service
-        query_protocol.deferred.addCallbacks(store_service_up, store_service_not_up)
+        query_protocol.deferred.addCallbacks(
+            store_service_up, store_service_not_up)
         return query_protocol.deferred
 
     def initialize_services(host):
@@ -155,8 +188,11 @@ def status(hosts=None, include_artefacts=True, use_cache_only=False, **kwargs):
         for settings in services:
             if type(settings) is str or not settings:
                 name = settings
-                settings = None
-                service_class = 'Service'
+                settings = services.get(settings, None)
+                if settings:
+                    service_class = settings.get('class', 'Service')
+                else:
+                    service_class = 'Service'
             else:
                 name = settings.keys()[0]
                 settings = settings[name]
@@ -171,7 +207,8 @@ def status(hosts=None, include_artefacts=True, use_cache_only=False, **kwargs):
                         service = clazz(host, name, settings)
                         break
             if not service:
-                host.logger.debug('%s not a standard service, searching class' % service_class)
+                host.logger.debug(
+                    '%s not a standard service, searching class' % service_class)
                 clazz = None
                 try:
                     host.logger.debug('fallback 1: checking loaded modules')
@@ -188,31 +225,40 @@ def status(hosts=None, include_artefacts=True, use_cache_only=False, **kwargs):
 
                 if not clazz:
                     try:
-                        host.logger.debug('fallback 2: trying to load module myself')
+                        host.logger.debug(
+                            'fallback 2: trying to load module myself')
                         clazz = get_class(service_class)
                     except Exception, e:
                         host.logger.debug(e)
                 if not clazz:
                     try:
-                        host.logger.debug('fallback 3: trying to lookup %s in legacies' % service_class)
+                        host.logger.debug(
+                            'fallback 3: trying to lookup %s in legacies' % service_class)
                         import legacies
-                        mapped_service_class = legacies.MAPPING_OLD_NEW_SERVICECLASSES.get(service_class, service_class)
+                        mapped_service_class = legacies.MAPPING_OLD_NEW_SERVICECLASSES.get(
+                            service_class, service_class)
                         clazz = get_class(mapped_service_class)
-                        host.logger.info('deprecation info: class %s was mapped to %s' % (service_class, mapped_service_class))
+                        host.logger.info(
+                            'deprecation info: class %s was mapped to %s' %
+                            (service_class, mapped_service_class))
                     except Exception, e:
                         host.logger.debug(e)
 
                 if not clazz:
-                    raise Exception('cannot find class %(service_class)s' % locals())
+                    raise Exception(
+                        'cannot find class %(service_class)s' % locals())
 
                 try:
                     service = clazz(host, name, settings)
                 except Exception, e:
                     host.logger.exception(e)
             if not service:
-                raise Exception('cannot instantiate class %(service_class)s' % locals())
+                raise Exception(
+                    'cannot instantiate class %(service_class)s' % locals())
             components[service.uri] = service
             service.fqdn = host.fqdn
+            service.needs = getattr(service, 'needs', set())
+            service.needs.add(host.uri)
             host.defined_services.append(service)
         return host
 
@@ -224,7 +270,8 @@ def status(hosts=None, include_artefacts=True, use_cache_only=False, **kwargs):
                     service.prepare(host)
                 if hasattr(service, 'get_local_service_collector'):
                     global local_service_collector
-                    local_service_collector = service.get_local_service_collector()
+                    local_service_collector = service.get_local_service_collector(
+                    )
                 q = query_local_service(service)
                 local_state.append(q)
 
@@ -237,7 +284,8 @@ def status(hosts=None, include_artefacts=True, use_cache_only=False, **kwargs):
     def initialize_artefacts(host):
         try:
             for version in getattr(host, 'current_artefacts', []):
-                artefact = yadtshell.components.Artefact(host, version, version)
+                artefact = yadtshell.components.Artefact(
+                    host, version, version)
                 artefact.state = yadtshell.settings.INSTALLED
                 components[artefact.uri] = artefact
         except TypeError:
@@ -246,10 +294,13 @@ def status(hosts=None, include_artefacts=True, use_cache_only=False, **kwargs):
 
         try:
             for version in getattr(host, 'current_artefacts', []):
-                uri = yadtshell.uri.create(yadtshell.settings.ARTEFACT, host.host, version)
-                artefact = components.get(uri, yadtshell.components.MissingComponent(uri))
+                uri = yadtshell.uri.create(
+                    yadtshell.settings.ARTEFACT, host.host, version)
+                artefact = components.get(
+                    uri, yadtshell.components.MissingComponent(uri))
                 artefact.revision = yadtshell.settings.CURRENT
-                current_uri = yadtshell.uri.create(yadtshell.settings.ARTEFACT, host.host, artefact.name + '/' + yadtshell.settings.CURRENT)
+                current_uri = yadtshell.uri.create(
+                    yadtshell.settings.ARTEFACT, host.host, artefact.name + '/' + yadtshell.settings.CURRENT)
                 components[uri] = artefact
                 components[current_uri] = artefact
         except TypeError:
@@ -257,7 +308,8 @@ def status(hosts=None, include_artefacts=True, use_cache_only=False, **kwargs):
 
         try:
             for version in getattr(host, 'next_artefacts', set()):
-                artefact = yadtshell.components.Artefact(host, version, version)
+                artefact = yadtshell.components.Artefact(
+                    host, version, version)
                 artefact.state = yadtshell.settings.INSTALLED
                 artefact.revision = yadtshell.settings.NEXT
                 components[artefact.uri] = artefact
@@ -265,10 +317,13 @@ def status(hosts=None, include_artefacts=True, use_cache_only=False, **kwargs):
             pass
         try:
             for version in getattr(host, 'next_artefacts', []):
-                uri = yadtshell.uri.create(yadtshell.settings.ARTEFACT, host.host, version)
-                artefact = components.get(uri, yadtshell.components.MissingComponent(uri))
+                uri = yadtshell.uri.create(
+                    yadtshell.settings.ARTEFACT, host.host, version)
+                artefact = components.get(
+                    uri, yadtshell.components.MissingComponent(uri))
                 artefact.revision = yadtshell.settings.NEXT
-                next_uri = yadtshell.uri.create(yadtshell.settings.ARTEFACT, host.host, artefact.name + '/' + yadtshell.settings.NEXT)
+                next_uri = yadtshell.uri.create(
+                    yadtshell.settings.ARTEFACT, host.host, artefact.name + '/' + yadtshell.settings.NEXT)
                 components[uri] = artefact
                 components[next_uri] = artefact
                 host.logger.debug('adding %(uri)s and %(next_uri)s' % locals())
@@ -317,18 +372,23 @@ def status(hosts=None, include_artefacts=True, use_cache_only=False, **kwargs):
                 except KeyError, ke:
                     logger.warning("unknown dependent key " + str(ke))
 
+        compute_dependency_scores(components)
+
+        def _open_component_file(component_type):
+            return open(os.path.join(yadtshell.settings.OUT_DIR, component_type), 'w')
+
         component_files = {
-                yadtshell.settings.ARTEFACT:   open(os.path.join(yadtshell.settings.OUT_DIR, 'artefacts'), 'w'),
-                yadtshell.settings.SERVICE:    open(os.path.join(yadtshell.settings.OUT_DIR, 'services'), 'w'),
-                yadtshell.settings.HOST:       open(os.path.join(yadtshell.settings.OUT_DIR, 'hosts'), 'w'),
-                }
+            yadtshell.settings.ARTEFACT: _open_component_file('artefacts'),
+            yadtshell.settings.SERVICE: _open_component_file('services'),
+            yadtshell.settings.HOST: _open_component_file('hosts'),
+        }
         for component in components.values():
             print >> component_files[component.type], component.uri
 
         for f in component_files.values():
             f.close()
 
-        f = open(os.path.join(yadtshell.settings.OUT_DIR, 'current_state.components'), "w")
+        f = _open_component_file('current_state.components')
         pickle.dump(components, f)
         f.close()
 
@@ -339,12 +399,16 @@ def status(hosts=None, include_artefacts=True, use_cache_only=False, **kwargs):
             for hostname in he.expand(grouped_hosts):
                 services = []
                 host = components['host://%s' % hostname]
-                for service in getattr(host, 'defined_services', []):
+
+                host_services = getattr(host, 'defined_services', [])
+                host_services.sort(key=lambda s: s.dependency_score)
+
+                for service in host_services:
                     services.append({
                         'uri': service.uri,
                         'name': service.name,
                         'state': service.state
-                        })
+                    })
 
                 artefacts = []
                 for artefact in sorted(getattr(host, 'handled_artefacts', [])):
@@ -353,16 +417,17 @@ def status(hosts=None, include_artefacts=True, use_cache_only=False, **kwargs):
                         'uri': 'artefact://%s/%s' % (hostname, name),
                         'name': name,
                         'current': version
-                        })
+                    })
 
                 host = {
-                        'name': hostname,
-                        'services': services,
-                        'artefacts': artefacts
-                        }
+                    'name': hostname,
+                    'services': services,
+                    'artefacts': artefacts
+                }
                 hosts.append(host)
             groups.append(hosts)
-        yadtshell.settings.ybc.sendFullUpdate(groups)
+        yadtshell.settings.ybc.sendFullUpdate(
+            groups, tracking_id=yadtshell.settings.tracking_id)
 
         status_line = yadtshell.util.get_status_line(components)
         logger.debug('\nstatus: %s' % status_line)
@@ -379,44 +444,26 @@ def status(hosts=None, include_artefacts=True, use_cache_only=False, **kwargs):
 
     def report_connection_error(failure):
         if failure.value.exitCode == 255:
-            logger.critical('ssh: cannot reach %s' % failure.value.component)
-            logger.info('passwordless ssh not configured? network problems?')
-            return yadtshell.twisted.SshFailure('ssh connect')
+            logger.critical(
+                'ssh: cannot reach %s\n\t passwordless ssh not configured? network problems?' %
+                failure.value.component)
+            return yadtshell.components.UnreachableHost(failure.value.component)
         return failure
 
     def notify_collector(ignored):
         global local_service_collector
         if local_service_collector:
-            logger.debug("collected services: %s " % ", ".join(local_service_collector.services))
+            logger.debug("collected services: %s " %
+                         ", ".join(local_service_collector.services))
             return local_service_collector.notify()
 
-    def restore_cached_state(component):
-        d = defer.Deferred()
-
-        state_file = os.path.join(yadtshell.settings.OUT_DIR, 'current_state_%s.yaml' % component)
-        logger.info('restoring cached data from %s' % state_file)
-
-        try:
-            f = open(state_file)
-            result = f.read()
-            f.close()
-        except IOError, e:
-            logger.warning(str(e))
-        host = yaml.load(result)
-        components[host.uri] = host
-        reactor.callLater(.1, d.callback, host)
-        return d
-
     pi = yadtshell.twisted.ProgressIndicator()
 
     deferreds = []
     for host in hosts:
-        if use_cache_only:
-            deferred = restore_cached_state(host)
-        else:
-            deferred = query_status(host, pi)
-            deferred.addErrback(report_connection_error)
-            deferred.addCallback(create_host)
+        deferred = query_status(host, pi)
+        deferred.addErrback(report_connection_error)
+        deferred.addCallback(create_host, components, Loader)
 
         deferred.addCallback(initialize_services)
         deferred.addCallback(add_local_state)
@@ -431,6 +478,7 @@ def status(hosts=None, include_artefacts=True, use_cache_only=False, **kwargs):
     dl.addCallback(notify_collector)
     dl.addCallback(build_unified_dependencies_tree)
     dl.addCallback(yadtshell.info, components=components)
-    dl.addErrback(yadtshell.twisted.report_error, logger.error, include_stacktrace=False)
+    dl.addErrback(yadtshell.twisted.report_error,
+                  logger.error, include_stacktrace=False)
 
     return dl
