@@ -74,9 +74,10 @@ class ProgressIndicator(object):
         return value
 
     def _render_compressed(self):
-        rendered = [self._render_value(self.progress.get(o)) for o in self.observables]
+        rendered = [self._render_value(self.progress.get(o))
+                    for o in self.observables]
         finished = [r for r in rendered if r not in self.rendered]
-        unfinished = [r for r in rendered if r in self.rendered]    # TODO mhh, could be one call only?!
+        unfinished = [r for r in rendered if r in self.rendered]
         finished_histo = dict((i, finished.count(i)) for i in set(finished))
         return ''.join(['%(count)i*%(key)s ' % locals() for key, count in finished_histo.iteritems()]) + ''.join([str(o) for o in unfinished])
 
@@ -90,9 +91,14 @@ class ProgressIndicator(object):
 
 
 class YadtProcessProtocol(protocol.ProcessProtocol):
-    def __init__(self, component, cmd, pi=None, out_log_level=logging.DEBUG, err_log_level=logging.WARN, log_prefix=''):
-        self.component = component
-        self.cmd = cmd
+    def __init__(self, component, cmd, pi=None, out_log_level=logging.DEBUG,
+                 err_log_level=logging.WARN, log_prefix='', wait_for_io=True):
+        try:
+            self.component = component.encode('ascii')
+        except AttributeError:
+            self.component = component
+        self.cmd = cmd.encode('ascii')
+        self.wait_for_io = wait_for_io
         self.data = ""
         self.pi = pi
         if not log_prefix:
@@ -109,34 +115,37 @@ class YadtProcessProtocol(protocol.ProcessProtocol):
             self.pi.update((self.cmd, self.component))
 
     def outReceived(self, data):
-        self.logger.debug("outReceived! with %d bytes!" % len(data))
         for line in data.splitlines():
-            self.logger.log(self.out_log_level, 'stdout: %s' % line)
+            self.logger.log(self.out_log_level,
+                            '{0}: {1}'.format(self.component, line))
         self.data = self.data + data
         if self.pi:
             self.pi.update((self.cmd, self.component))
 
     def errReceived(self, data):
-        self.logger.debug("errReceived! with %d bytes!" % len(data))
         for line in data.splitlines():
-            self.logger.log(self.err_log_level, 'stderr: %s' % line)
+            self.logger.log(self.err_log_level,
+                            '{0} {1} stderr: {2}'.format(self.component,
+                                                         self.cmd,
+                                                         line))
         if self.pi:
             self.pi.update((self.cmd, self.component))
 
-    def inConnectionLost(self):
-        self.logger.debug("inConnectionLost! stdin is closed! (we probably did it)")
-
-    def outConnectionLost(self):
-        self.logger.debug("outConnectionLost! The child closed their stdout!")
-
-    def errConnectionLost(self):
-        self.logger.debug("errConnectionLost! The child closed their stderr.")
-
     def processExited(self, reason):
-        self.logger.debug("processExited, exit code %s" % str(reason.value.exitCode))
+        self.logger.debug("%s@%s exited, exit code %s" % (self.cmd,
+                                                          self.component,
+                                                          str(reason.value.exitCode)))
+        if not self.wait_for_io:
+            self.finish(reason)
 
     def processEnded(self, reason):
-        self.logger.debug("status received, exit code %s" % str(reason.value.exitCode))
+        self.logger.debug("%s@%s ended, exit code %s" % (self.cmd,
+                                                         self.component,
+                                                         str(reason.value.exitCode)))
+        if self.wait_for_io:
+            self.finish(reason)
+
+    def finish(self, reason):
         self.exitcode = reason.value.exitCode
         if self.pi:
             self.pi.update((self.cmd, self.component), str(self.exitcode))
@@ -166,6 +175,13 @@ def stop_and_return(return_code):
     stop_yadt()
 
 
+def _determine_issued_command(failure):
+    try:
+        return '%s@' % failure.value.orig_protocol.cmd
+    except AttributeError:
+        return ''
+
+
 def report_error(failure, line_fun=None, include_stacktrace=True):
     if line_fun is None:
         def line_fun(line):
@@ -173,7 +189,9 @@ def report_error(failure, line_fun=None, include_stacktrace=True):
     if isinstance(failure, SshFailure):
         return failure
     if hasattr(failure.value, 'component'):
-        line_fun('%s: %s' % (failure.value.component, failure.getErrorMessage()))
+        line_fun('%s%s: %s' % (_determine_issued_command(failure),
+                               failure.value.component,
+                               failure.getErrorMessage()))
     else:
         line_fun(failure.getErrorMessage())
         if include_stacktrace:
