@@ -16,19 +16,24 @@
 #   You should have received a copy of the GNU General Public License
 #   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
+from __future__ import absolute_import
+
 import logging
 import os.path
 import pwd
 import socket
-import subprocess
 import sys
 import yaml
+import shlex
 
 from time import localtime, strftime
+from twisted.internet import defer, reactor
 
 import yadtshell.settings
 import yadtshell.components
 
+from yadtshell.constants import STANDALONE_SERVICE_RANK
+from yadtshell.validation import ServiceDefinitionValidator
 
 logger = logging.getLogger('util')
 
@@ -80,19 +85,25 @@ def dump_plan(flavor, plan):
         yaml.dump(plan, f)
     f.close()
 
+
 def dump_action_plan(flavor, plan):
     dump_plan(flavor + '-action', plan)
 
 
+def current_state():
+    return os.path.join(yadtshell.settings.OUT_DIR, 'current_state.components')
+
+
 def restore_current_state():
     try:
-        return restore(os.path.join(yadtshell.settings.OUT_DIR, 'current_state.components'))
+        return restore(current_state())
     except IOError:
         logger.warning('no current state stored, try "status" first')
         sys.exit(1)
 
+
 def get_mtime_of_current_state():
-    return os.path.getmtime(os.path.join(yadtshell.settings.OUT_DIR, 'current_state.components'))  # TODO combine this with prev method
+    return os.path.getmtime(current_state())
 
 
 def is_up(state):
@@ -101,7 +112,6 @@ def is_up(state):
 
 def not_up(state):
     return not is_up(state)
-    #return state in [None, settings.UNKNOWN, settings.DOWN, settings.MISSING, settings.UPDATE_NEEDED]
 
 
 def render_state(state, just='left', width=10):
@@ -151,6 +161,7 @@ def get_locking_user_info():
             "pid": pid,
             }
 
+
 def get_yaml(adict):
     return yaml.dump(adict, default_flow_style=False)
 
@@ -196,29 +207,89 @@ def get_status_line(components):
 def start_ssh_multiplexed(hosts=None):
     if not hosts:
         hosts = yadtshell.settings.TARGET_SETTINGS['hosts']
-    for host in hosts:
-        ssh_check_cmds = ['ssh',
-            '-o', 'ControlPath=%s' % yadtshell.settings.SSH_CONTROL_PATH,
-            '-O', 'check', host]
-        ssh_check_call = subprocess.Popen(ssh_check_cmds, stderr=subprocess.PIPE)
-        ssh_check_call.communicate()
-        if ssh_check_call.returncode != 0:
-            start_multiplexing_call = ['ssh', '-fN', '-o', 'ControlPath=%s' % yadtshell.settings.SSH_CONTROL_PATH, '-o', 'ControlMaster=yes', host]
-            logger.debug(' '.join(start_multiplexing_call))
-            is_started = subprocess.call(start_multiplexing_call)
-            logger.debug(is_started)
-            logger.debug('multiplexed ssh connection to %(host)s created' % locals())
+
+    def start_ssh(protocol, host):
+        logger.debug('start_ssh %s' % host)
+        start_multiplexing_call = shlex.split(
+            '%s -fN -o ControlMaster=yes %s' % (yadtshell.settings.SSH, host))
+        p = yadtshell.twisted.YadtProcessProtocol(
+            host, 'start_ssh', wait_for_io=False)
+        p.deferred = defer.Deferred()
+        logger.debug('cmd: %s' % start_multiplexing_call)
+        reactor.spawnProcess(
+            p, start_multiplexing_call[0], start_multiplexing_call, None)
+        return protocol
+
+    def check_ssh(host):
+        ssh_check_cmds = shlex.split(
+            '%s -O check %s' % (yadtshell.settings.SSH, host))
+        p = yadtshell.twisted.YadtProcessProtocol(host, 'check_ssh')
+        p.deferred = defer.Deferred()
+        logger.debug('cmd: %s' % ssh_check_cmds)
+        reactor.spawnProcess(p, ssh_check_cmds[0], ssh_check_cmds, None)
+        p.deferred.addErrback(start_ssh, host)
+        return p.deferred
+
+    return defer.DeferredList([check_ssh(host) for host in hosts])
 
 
 def stop_ssh_multiplexed(ignored, hosts=None):
+
+    def stop_ssh(host):
+        ssh_stop_cmds = shlex.split(
+            '%s -O exit %s' % (yadtshell.settings.SSH, host))
+        p = yadtshell.twisted.YadtProcessProtocol(host, 'stop_ssh')
+        p.deferred = defer.Deferred()
+        logger.debug('cmd: %s' % ssh_stop_cmds)
+        reactor.spawnProcess(
+            p, ssh_stop_cmds[0], ssh_stop_cmds, None, childFDs={2: 3})
+        return p.deferred
+
     if not hosts:
         hosts = yadtshell.settings.TARGET_SETTINGS['hosts']
-    for host in hosts:
-        logger.debug('multiplexed ssh connections to %(host)s removed' % locals())
-        ssh_check_cmds = ['ssh',
-            '-o', 'ControlPath=%s' % yadtshell.settings.SSH_CONTROL_PATH,
-            '-O', 'exit', host]
-        #ignore = subprocess.call(ssh_check_call)
-        stop_multiplexing_call = subprocess.Popen(ssh_check_cmds, stderr=subprocess.PIPE)
-        stop_multiplexing_call.communicate()
-    return ignored
+
+    dl = defer.DeferredList([stop_ssh(host) for host in hosts])
+
+    dl.addCallback(lambda _: ignored)
+    return dl
+
+
+def inbound_deps_on_same_host(service, components):
+    inbound_services = [
+        s for s in service.needed_by if 'service://%s' % service.host in s]
+    for dependent_service in service.needed_by:
+        inbound_services.extend(
+            inbound_deps_on_same_host(components[dependent_service], components))
+    return inbound_services
+
+
+def outbound_deps_on_same_host(service, components):
+    needed_services = [
+        s for s in service.needs if 'service://%s' % service.host in s]
+    outbound_services = needed_services
+    for needed_service in [s for s in service.needs if 'service://%s' % service.host in s]:
+        outbound_services.extend(
+            outbound_deps_on_same_host(components[needed_service], components))
+    return outbound_services
+
+
+def compute_dependency_scores(components):
+    servicedefs = dict((component.uri, component)
+                       for component in components.values() if isinstance(component, yadtshell.components.Service))
+
+    #  we cannot compute the dependency scores if there is a service cycle
+    t = ServiceDefinitionValidator(servicedefs)
+    t.assert_no_cycles_present()
+
+    for service, servicedef in servicedefs.iteritems():
+        outbound_edges = len(
+            outbound_deps_on_same_host(servicedef, components))
+        inbound_edges = len(inbound_deps_on_same_host(servicedef, components))
+        if outbound_edges == inbound_edges == 0:
+            servicedef.dependency_score = STANDALONE_SERVICE_RANK
+        else:
+            servicedef.dependency_score = inbound_edges - outbound_edges
+
+
+def calculate_max_tries_for_interval_and_delay(interval, delay):
+    return (interval + delay - 1) / delay
